@@ -5,6 +5,7 @@
 
 #![recursion_limit = "128"]
 #![cfg_attr(nightly, feature(proc_macro_diagnostic))]
+#![allow(clippy::let_and_return)]
 
 extern crate proc_macro;
 
@@ -13,11 +14,12 @@ use proc_macro2::{Span, TokenStream};
 use quote::{quote, ToTokens, TokenStreamExt};
 use std::collections::HashMap;
 use std::fmt::Write;
+use syn::punctuated::Punctuated;
 #[cfg(nightly)]
 use syn::spanned::Spanned;
 use syn::Token;
 use syn::{parse_macro_input, parse_quote};
-use syn::{GenericParam, Ident, Type, TypeParam, TypePath};
+use syn::{GenericParam, Ident, Type, TypeParam, TypePath, WhereClause, WherePredicate};
 
 mod args;
 mod layout;
@@ -79,7 +81,10 @@ impl<'a> ToTokens for SubstTyGenerics<'a> {
 /// Macro to derive widget traits
 ///
 /// See the [`kas::macros`](../../kas/macros/index.html) module documentation.
-#[proc_macro_derive(Widget, attributes(widget_core, widget, layout, handler, layout_data))]
+#[proc_macro_derive(
+    Widget,
+    attributes(handler, layout, layout_data, widget, widget_core, widget_derive)
+)]
 pub fn derive(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
     let mut ast = parse_macro_input!(input as syn::DeriveInput);
 
@@ -87,11 +92,24 @@ pub fn derive(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
         Ok(w) => w,
         Err(err) => return err.to_compile_error().into(),
     };
+
+    let derive_inner = args.core_data.is_none();
+    let opt_inner = args.inner.as_ref().map(|(inner, _)| inner.clone());
+
     let (impl_generics, ty_generics, where_clause) = ast.generics.split_for_impl();
     let name = &ast.ident;
     let widget_name = name.to_string();
 
-    let core_data = args.core_data;
+    let (core_data, core_data_mut) = args
+        .core_data
+        .map(|cd| (quote! { &self.#cd }, quote! { &mut self.#cd }))
+        .unwrap_or_else(|| {
+            let inner = opt_inner.as_ref().unwrap();
+            (
+                quote! { self.#inner.core_data() },
+                quote! { self.#inner.core_data_mut() },
+            )
+        });
     let mut toks = quote! {
         impl #impl_generics kas::WidgetCore
             for #name #ty_generics #where_clause
@@ -100,11 +118,11 @@ pub fn derive(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
             fn as_any_mut(&mut self) -> &mut dyn std::any::Any { self }
 
             fn core_data(&self) -> &kas::CoreData {
-                &self.#core_data
+                #core_data
             }
 
             fn core_data_mut(&mut self) -> &mut kas::CoreData {
-                &mut self.#core_data
+                #core_data_mut
             }
 
             fn widget_name(&self) -> &'static str {
@@ -116,7 +134,30 @@ pub fn derive(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
         }
     };
 
-    if args.widget.children {
+    if derive_inner {
+        let inner = opt_inner.as_ref().unwrap();
+        toks.append_all(quote! {
+            impl #impl_generics kas::WidgetChildren
+                for #name #ty_generics #where_clause
+            {
+                fn first_id(&self) -> kas::WidgetId {
+                    self.#inner.first_id()
+                }
+                fn record_first_id(&mut self, id: WidgetId) {
+                    self.#inner.record_first_id(id);
+                }
+                fn num_children(&self) -> usize {
+                    self.#inner.num_children()
+                }
+                fn get_child(&self, index: usize) -> Option<&dyn kas::WidgetConfig> {
+                    self.#inner.get_child(index)
+                }
+                fn get_child_mut(&mut self, index: usize) -> Option<&mut dyn kas::WidgetConfig> {
+                    self.#inner.get_child_mut(index)
+                }
+            }
+        });
+    } else if args.widget.children {
         let first_id = if args.children.is_empty() {
             quote! { self.id() }
         } else {
@@ -182,7 +223,39 @@ pub fn derive(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
         });
     }
 
-    if let Some(ref layout) = args.layout {
+    if derive_inner {
+        let inner = opt_inner.as_ref().unwrap();
+        toks.append_all(quote! {
+            impl #impl_generics kas::Layout
+                    for #name #ty_generics #where_clause
+            {
+                #[inline]
+                fn size_rules(&mut self, size_handle: &mut dyn SizeHandle, axis: AxisInfo) -> SizeRules {
+                    self.#inner.size_rules(size_handle, axis)
+                }
+                #[inline]
+                fn set_rect(&mut self, mgr: &mut Manager, rect: Rect, align: AlignHints) {
+                    self.#inner.set_rect(mgr, rect, align);
+                }
+                #[inline]
+                fn translation(&self, child_index: usize) -> Offset {
+                    self.#inner.translation(child_index)
+                }
+                #[inline]
+                fn spatial_nav(&self, reverse: bool, from: Option<usize>) -> Option<usize> {
+                    self.#inner.spatial_nav(reverse, from)
+                }
+                #[inline]
+                fn find_id(&self, coord: Coord) -> Option<WidgetId> {
+                    self.#inner.find_id(coord)
+                }
+                #[inline]
+                fn draw(&self, draw_handle: &mut dyn DrawHandle, mgr: &ManagerState, disabled: bool) {
+                    self.#inner.draw(draw_handle, mgr, disabled);
+                }
+            }
+        });
+    } else if let Some(ref layout) = args.layout {
         match layout::data_type(&args.children, layout) {
             Ok(dt) => toks.append_all(quote! {
                 impl #impl_generics kas::LayoutData
@@ -219,7 +292,7 @@ pub fn derive(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
             .params
             .into_pairs()
             .filter(|pair| match pair.value() {
-                &GenericParam::Type(TypeParam { ref ident, .. }) => !subs.contains_key(ident),
+                GenericParam::Type(TypeParam { ref ident, .. }) => !subs.contains_key(ident),
                 _ => true,
             })
             .collect();
@@ -258,36 +331,63 @@ pub fn derive(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
 
         if handler.handle {
             let msg = handler.msg;
+            let handle = if derive_inner {
+                let inner = opt_inner.as_ref().unwrap();
+                quote! {
+                    #[inline]
+                    fn activation_via_press(&self) -> bool {
+                        self.#inner.activation_via_press()
+                    }
+                    #[inline]
+                    fn handle(&mut self, mgr: &mut Manager, event: Event) -> Response<Self::Msg> {
+                        self.#inner.handle(mgr, event)
+                    }
+                }
+            } else {
+                quote! {}
+            };
             toks.append_all(quote! {
                 impl #impl_generics kas::event::Handler
                         for #name #ty_generics #where_clause
                 {
                     type Msg = #msg;
+                    #handle
                 }
             });
         }
 
         if handler.send {
-            let mut ev_to_num = TokenStream::new();
-            for child in args.children.iter() {
-                let ident = &child.ident;
-                let handler = if let Some(ref h) = child.args.handler {
-                    quote! { r.try_into().unwrap_or_else(|msg| self.#h(mgr, msg)) }
-                } else {
-                    quote! { r.into() }
-                };
-                ev_to_num.append_all(quote! {
-                    if id <= self.#ident.id() {
-                        let r = self.#ident.send(mgr, id, event);
-                        #handler
-                    } else
-                });
-            }
+            let send_impl = if derive_inner {
+                let inner = opt_inner.as_ref().unwrap();
+                quote! { self.#inner.send(mgr, id, event) }
+            } else {
+                let mut ev_to_num = TokenStream::new();
+                for child in args.children.iter() {
+                    let ident = &child.ident;
+                    let handler = if let Some(ref h) = child.args.handler {
+                        quote! {
+                            r.try_into().unwrap_or_else(|msg| {
+                                log::trace!(
+                                    "Received by {} from {}: {:?}",
+                                    self.id(),
+                                    id,
+                                    kas::util::TryFormat(&msg)
+                                );
+                                self.#h(mgr, msg)
+                            })
+                        }
+                    } else {
+                        quote! { r.into() }
+                    };
+                    ev_to_num.append_all(quote! {
+                        if id <= self.#ident.id() {
+                            let r = self.#ident.send(mgr, id, event);
+                            #handler
+                        } else
+                    });
+                }
 
-            let send = quote! {
-                fn send(&mut self, mgr: &mut kas::event::Manager, id: kas::WidgetId, event: kas::event::Event)
-                -> kas::event::Response<Self::Msg>
-                {
+                quote! {
                     use kas::{WidgetCore, event::Response};
                     if self.is_disabled() {
                         return Response::Unhandled;
@@ -304,7 +404,15 @@ pub fn derive(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
                 impl #impl_generics kas::event::SendEvent
                         for #name #ty_generics #where_clause
                 {
-                    #send
+                    fn send(
+                        &mut self,
+                        mgr: &mut kas::event::Manager,
+                        id: kas::WidgetId,
+                        event: kas::event::Event
+                    ) -> kas::event::Response<Self::Msg>
+                    {
+                        #send_impl
+                    }
                 }
             });
         }
@@ -314,6 +422,110 @@ pub fn derive(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
         });
     }
 
+    if let Some((member, ty)) = args.inner {
+        if args.derive.deref {
+            toks.append_all(quote! {
+                impl #impl_generics std::ops::Deref for #name #ty_generics #where_clause {
+                    type Target = #ty;
+                    #[inline]
+                    fn deref(&self) -> &Self::Target {
+                        &self.#member
+                    }
+                }
+            });
+        }
+
+        if args.derive.deref_mut {
+            toks.append_all(quote! {
+                impl #impl_generics std::ops::DerefMut for #name #ty_generics #where_clause {
+                    #[inline]
+                    fn deref_mut(&mut self) -> &mut Self::Target {
+                        &mut self.#member
+                    }
+                }
+            });
+        }
+
+        let extended_where_clause = move |pred: WherePredicate| {
+            if let Some(clause) = where_clause {
+                let mut clauses: WhereClause = (*clause).clone();
+                clauses.predicates.push_punct(Default::default());
+                clauses.predicates.push_value(pred);
+                clauses
+            } else {
+                let mut predicates = Punctuated::new();
+                predicates.push_value(pred);
+                WhereClause {
+                    where_token: Default::default(),
+                    predicates,
+                }
+            }
+        };
+
+        if args.derive.has_bool {
+            let wc = extended_where_clause(parse_quote! { #ty: kas::class::HasBool });
+            toks.append_all(quote! {
+                impl #impl_generics kas::class::HasBool for #name #ty_generics #wc {
+                    #[inline]
+                    fn get_bool(&self) -> bool {
+                        self.#member.get_bool()
+                    }
+
+                    #[inline]
+                    fn set_bool(&mut self, state: bool) -> kas::TkAction {
+                        self.#member.set_bool(state)
+                    }
+                }
+            });
+        }
+
+        if args.derive.has_str {
+            let wc = extended_where_clause(parse_quote! { #ty: kas::class::HasStr });
+            toks.append_all(quote! {
+                impl #impl_generics kas::class::HasStr for #name #ty_generics #wc {
+                    #[inline]
+                    fn get_str(&self) -> &str {
+                        self.#member.get_str()
+                    }
+
+                    #[inline]
+                    fn get_string(&self) -> String {
+                        self.#member.get_string()
+                    }
+                }
+            });
+        }
+
+        if args.derive.has_string {
+            let wc = extended_where_clause(parse_quote! { #ty: kas::class::HasString });
+            toks.append_all(quote! {
+                impl #impl_generics kas::class::HasString for #name #ty_generics #wc {
+                    #[inline]
+                    fn set_str(&mut self, text: &str) -> kas::TkAction {
+                        self.#member.set_str(text)
+                    }
+
+                    #[inline]
+                    fn set_string(&mut self, text: String) -> kas::TkAction {
+                        self.#member.set_string(text)
+                    }
+                }
+            });
+        }
+
+        if args.derive.set_accel {
+            let wc = extended_where_clause(parse_quote! { #ty: kas::class::SetAccel });
+            toks.append_all(quote! {
+                impl #impl_generics kas::class::SetAccel for #name #ty_generics #wc {
+                    #[inline]
+                    fn set_accel_string(&mut self, accel: AccelString) -> kas::TkAction {
+                        self.#member.set_accel_string(accel)
+                    }
+                }
+            });
+        }
+    }
+
     toks.into()
 }
 
@@ -429,7 +641,7 @@ pub fn make_widget(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
 
                 for item in body {
                     match item {
-                        &syn::ImplItem::Type(syn::ImplItemType {
+                        syn::ImplItem::Type(syn::ImplItemType {
                             ref ident, ref ty, ..
                         }) if *ident == msg_ident => {
                             msg = Some(ty.clone());
@@ -508,7 +720,7 @@ pub fn make_widget(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
                                 return quote! {}.into(); // exit after emitting error
                             }
                         } else {
-                            name_buf.push_str("R");
+                            name_buf.push('R');
                             let tyr = Ident::new(&name_buf, Span::call_site());
                             handler
                                 .generics
